package api

import (
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"path/filepath"

	"github.com/meshplus/bitxhub-kit/crypto/asym"

	"github.com/gin-gonic/gin"
	"github.com/link33/sidecar/cmd/sidecar/client"
	"github.com/link33/sidecar/internal/peermgr"
	"github.com/link33/sidecar/internal/repo"
	"github.com/link33/sidecar/model/pb"
	appchainmgr "github.com/meshplus/bitxhub-core/appchain-mgr"
	"github.com/sirupsen/logrus"
)

type Server struct {
	router  *gin.Engine
	config  *repo.Config
	logger  logrus.FieldLogger
	ctx     context.Context
	cancel  context.CancelFunc
	peerMgr peermgr.PeerManager
}

type response struct {
	Data []byte `json:"data"`
}

func NewServer(config *repo.Config, peerMgr peermgr.PeerManager, logger logrus.FieldLogger) *Server {
	ctx, cancel := context.WithCancel(context.Background())
	gin.SetMode(gin.ReleaseMode)
	router := gin.New()
	return &Server{
		router:  router,
		config:  config,
		logger:  logger,
		ctx:     ctx,
		cancel:  cancel,
		peerMgr: peerMgr,
	}
}

func (g *Server) Start() error {
	g.router.Use(gin.Recovery())
	v1 := g.router.Group("/v1")
	{
		v1.POST(client.RegisterAppchainUrl, g.registerAppchain)
		v1.POST(client.UpdateAppchainUrl, g.updateAppchain)
		v1.GET(client.GetAppchainUrl, g.getAppchain)
	}

	go func() {
		go func() {
			err := g.router.Run(fmt.Sprintf(":%d", g.config.Port.Http))
			if err != nil {
				panic(err)
			}
		}()
		<-g.ctx.Done()
	}()
	return nil
}

func (g *Server) Stop() error {
	g.cancel()
	g.logger.Infoln("gin service stop")
	return nil
}

func (g *Server) updateAppchain(c *gin.Context) {
	g.sendAppchain(c, pb.Message_APPCHAIN_UPDATE)
}

func (g *Server) registerAppchain(c *gin.Context) {
	g.sendAppchain(c, pb.Message_APPCHAIN_REGISTER)
}

func (g *Server) sendAppchain(c *gin.Context, appchainType pb.Message_Type) {
	// target pier id
	pierID := c.Query("pier_id")
	var res pb.Response
	var appchain appchainmgr.Appchain
	if err := c.BindJSON(&appchain); err != nil {
		res.Data = []byte(err.Error())
		c.JSON(http.StatusBadRequest, res)
		return
	}
	data, err := json.Marshal(appchain)
	if err != nil {
		g.logger.Errorln(err)
	}

	msg := pb.Msg(appchainType, true, data)
	ackMsg, err := g.peerMgr.SendByID(pierID, msg)
	if err != nil {
		res.Data = []byte(err.Error())
		c.JSON(http.StatusInternalServerError, res)
		return
	}

	g.handleAckAppchain(c, ackMsg)
	panic("implement me")
}

func (g *Server) handleAckAppchain(c *gin.Context, msg *pb.Message) {
	app := appchainmgr.Appchain{}
	if err := json.Unmarshal(msg.Payload.Data, &app); err != nil {
		g.logger.Error(err)
		return
	}

	res := &response{}
	switch msg.Type {
	case pb.Message_APPCHAIN_REGISTER:
		res.Data = []byte(fmt.Sprintf("appchain register successfully, id is %s\n", app.ID))
	case pb.Message_APPCHAIN_UPDATE:
		res.Data = []byte(fmt.Sprintf("appchain update successfully, id is %s\n", app.ID))
	case pb.Message_APPCHAIN_GET:
		res.Data = msg.Payload.Data
	}

	c.JSON(http.StatusOK, res)
}

func (g *Server) getAppchain(c *gin.Context) {
	res := &response{}

	// target pier id
	selfPierID, err := g.getSelfPierID(c)
	if err != nil {
		c.JSON(http.StatusInternalServerError, err.Error())
		return
	}
	targetPierID := c.Query("pier_id")
	appchain := &appchainmgr.Appchain{
		ID: selfPierID,
	}

	data, err := json.Marshal(appchain)
	if err != nil {
		c.JSON(http.StatusInternalServerError, err.Error())
		return
	}
	msg := pb.Msg(pb.Message_APPCHAIN_GET, true, data)

	ackMsg, err := g.peerMgr.SendByID(targetPierID, msg)
	if err != nil {
		res.Data = []byte(err.Error())
		c.JSON(http.StatusInternalServerError, res)
		return
	}

	g.handleAckAppchain(c, ackMsg)
}

func (g *Server) getSelfPierID(ctx *gin.Context) (string, error) {
	repoRoot := g.config.RepoRoot
	keyPath := filepath.Join(repoRoot, "key.json")

	privKey, err := asym.RestorePrivateKey(keyPath, "link33")
	if err != nil {
		return "", err
	}

	address, err := privKey.PublicKey().Address()
	if err != nil {
		return "", err
	}

	return address.String(), nil
}
